@@ -4,12 +4,11 @@
  +-------------------------------------------------------------------------+
  | Enigma Plugin for Roundcube                                             |
  |                                                                         |
- | Copyright (C) 2010-2015 The Roundcube Dev Team                          |
+ | Copyright (C) The Roundcube Dev Team                                    |
  |                                                                         |
  | Licensed under the GNU General Public License version 3 or              |
  | any later version with exceptions for skins & plugins.                  |
  | See the README file for a full license statement.                       |
- |                                                                         |
  +-------------------------------------------------------------------------+
  | Author: Aleksander Machniak <alec@alec.pl>                              |
  +-------------------------------------------------------------------------+
@@ -21,7 +20,7 @@
  */
 class enigma extends rcube_plugin
 {
-    public $task = 'mail|settings';
+    public $task = 'mail|settings|cli';
     public $rc;
     public $engine;
     public $ui;
@@ -43,6 +42,7 @@ class enigma extends rcube_plugin
             $this->add_hook('message_body_prefix', array($this, 'status_message'));
 
             $this->register_action('plugin.enigmaimport', array($this, 'import_file'));
+            $this->register_action('plugin.enigmakeys', array($this, 'preferences_ui'));
 
             // load the Enigma plugin configuration
             $this->load_config();
@@ -74,6 +74,7 @@ class enigma extends rcube_plugin
             $this->add_hook('preferences_sections_list', array($this, 'preferences_sections_list'));
             $this->add_hook('preferences_list', array($this, 'preferences_list'));
             $this->add_hook('preferences_save', array($this, 'preferences_save'));
+            $this->add_hook('identity_form', array($this, 'identity_form'));
 
             // register handler for keys/certs management
             $this->register_action('plugin.enigmakeys', array($this, 'preferences_ui'));
@@ -84,6 +85,11 @@ class enigma extends rcube_plugin
             if (empty($_REQUEST['_framed']) || strpos($this->rc->action, 'plugin.enigma') === 0) {
                 $this->ui->add_css();
             }
+
+            $this->password_handler();
+        }
+        else if ($this->rc->task == 'cli') {
+            $this->add_hook('user_delete_commit', array($this, 'user_delete'));
         }
 
         $this->add_hook('refresh', array($this, 'refresh'));
@@ -417,6 +423,61 @@ class enigma extends rcube_plugin
         $this->ui->init();
     }
 
+    /**
+     * Handler for 'identity_form' plugin hook.
+     *
+     * This will list private keys matching this identity
+     * and add a link to enigma key management action.
+     *
+     * @param array Original parameters
+     *
+     * @return array Modified parameters
+     */
+    function identity_form($p)
+    {
+        if (isset($p['form']['encryption']) && !empty($p['record']['identity_id'])) {
+            $content = '';
+
+            // find private keys for this identity
+            if ($p['record']['email']) {
+                $listing = array();
+                $engine  = $this->load_engine();
+                $keys    = (array)$engine->list_keys($p['record']['email']);
+
+                foreach ($keys as $key) {
+                    if ($key->get_type() === enigma_key::TYPE_KEYPAIR) {
+                        $listing[] = html::tag('li', null,
+                            html::tag('strong', 'uid', html::quote($key->id))
+                            . ' ' . html::tag('span', 'identity', html::quote($key->name))
+                        );
+                    }
+                }
+
+                if (count($listing)) {
+                    $content .= html::p(null, $this->gettext(array('name' => 'identitymatchingprivkeys', 'vars' => array('nr' => count($listing)))));
+                    $content .= html::tag('ul', 'keylist', implode("\n", $listing));
+                } else {
+                    $content .= html::p(null, $this->gettext('identitynoprivkeys'));
+                }
+            }
+
+            // add button linking to enigma key management
+            $button_attr = array(
+                'class'  => 'button',
+                'href'   => $this->rc->url(array('action' => 'plugin.enigmakeys')),
+                'target' => '_parent',
+            );
+            $content .= html::p(null, html::a($button_attr, $this->gettext('managekeys')));
+
+            // rename class to avoid Mailvelope key management to kick in
+            $p['form']['encryption']['attrs'] = array('class' => 'enigma-identity-encryption');
+            // fill fieldset content with our stuff
+            $p['form']['encryption']['content'] = html::div('identity-encryption-block', $content);
+        }
+
+        return $p;
+    }
+
     /**
      * Handler for message_body_prefix hook.
      * Called for every displayed (content) part of the message.
@@ -508,4 +569,16 @@ class enigma extends rcube_plugin
 
         return $p;
     }
+
+    /**
+     * Handle delete_user_commit hook
+     */
+    function user_delete($p)
+    {
+        $this->load_engine();
+
+        $p['abort'] = $p['abort'] || !$this->engine->delete_user_data($p['username']);
+
+        return $p;
+    }
 }
