@@ -17,11 +17,10 @@
 
 class rcube_kpasswd_password
 {
-    public function save($currpass, $newpass)
+    public function save($currpass, $newpass, $username)
     {
         $bin      = rcmail::get_instance()->config->get('password_kpasswd_cmd', '/usr/bin/kpasswd');
-        $username = $_SESSION['username'];
-        $cmd      = $bin . ' "' . $username . '" 2>&1';
+        $cmd      = $bin . ' "' . escapeshellarg($username) . '" 2>&1';
 
         $handle = popen($cmd, "w");
         fwrite($handle, $currpass."\n");
@@ -31,14 +30,13 @@ class rcube_kpasswd_password
         if (pclose($handle) == 0) {
             return PASSWORD_SUCCESS;
         }
-        else {
-            rcube::raise_error(array(
+
+        rcube::raise_error(array(
                 'code' => 600,
                 'type' => 'php',
                 'file' => __FILE__, 'line' => __LINE__,
                 'message' => "Password plugin: Unable to execute $cmd"
-                ), true, false);
-        }
+            ), true, false);
 
         return PASSWORD_ERROR;
     }
