@@ -3,8 +3,9 @@
 /**
  +-----------------------------------------------------------------------+
  | This file is part of the Roundcube Webmail client                     |
- | Copyright (C) 2005-2016, The Roundcube Dev Team                       |
- | Copyright (C) 2011-2016, Kolab Systems AG                             |
+ |                                                                       |
+ | Copyright (C) The Roundcube Dev Team                                  |
+ | Copyright (C) Kolab Systems AG                                        |
  |                                                                       |
  | Licensed under the GNU General Public License version 3 or            |
  | any later version with exceptions for skins & plugins.                |
@@ -23,8 +24,6 @@
  *
  * @package    Framework
  * @subpackage Storage
- * @author     Thomas Bruederli <roundcube@gmail.com>
- * @author     Aleksander Machniak <alec@alec.pl>
  */
 class rcube_mime
 {
@@ -84,16 +83,21 @@ class rcube_mime
     /**
      * Split an address list into a structured array list
      *
-     * @param string  $input    Input string
-     * @param int     $max      List only this number of addresses
-     * @param boolean $decode   Decode address strings
-     * @param string  $fallback Fallback charset if none specified
-     * @param boolean $addronly Return flat array with e-mail addresses only
+     * @param string|array $input    Input string (or list of strings)
+     * @param int          $max      List only this number of addresses
+     * @param boolean      $decode   Decode address strings
+     * @param string       $fallback Fallback charset if none specified
+     * @param boolean      $addronly Return flat array with e-mail addresses only
      *
      * @return array Indexed list of addresses
      */
     static function decode_address_list($input, $max = null, $decode = true, $fallback = null, $addronly = false)
     {
+        // A common case when the same header is used many times in a mail message
+        if (is_array($input)) {
+            $input = implode(', ', $input);
+        }
+
         $a   = self::parse_address_list($input, $decode, $fallback);
         $out = array();
         $j   = 0;
@@ -706,9 +710,8 @@ class rcube_mime
     {
         static $mime_ext = array();
 
-        $mime_type  = null;
-        $config     = rcube::get_instance()->config;
-        $mime_magic = $config->get('mime_magic');
+        $mime_type = null;
+        $config    = rcube::get_instance()->config;
 
         if (!$skip_suffix && empty($mime_ext)) {
             foreach ($config->resolve_paths('mimetypes.php') as $fpath) {
@@ -725,6 +728,7 @@ class rcube_mime
 
         // try fileinfo extension if available
         if (!$mime_type && function_exists('finfo_open')) {
+            $mime_magic = $config->get('mime_magic');
             // null as a 2nd argument should be the same as no argument
             // this however is not true on all systems/versions
             if ($mime_magic) {
@@ -735,10 +739,8 @@ class rcube_mime
             }
 
             if ($finfo) {
-                if ($is_stream)
-                    $mime_type = finfo_buffer($finfo, $path);
-                else
-                    $mime_type = finfo_file($finfo, $path);
+                $func      = $is_stream ? 'finfo_buffer' : 'finfo_file';
+                $mime_type = $func($finfo, $path, FILEINFO_MIME_TYPE);
                 finfo_close($finfo);
             }
         }
@@ -752,11 +754,6 @@ class rcube_mime
         if (!$mime_type) {
             $mime_type = $failover;
         }
-        else {
-            // Sometimes (PHP-5.3?) content-type contains charset definition,
-            // Remove it (#1487122) also "charset=binary" is useless
-            $mime_type = array_shift(preg_split('/[; ]/', $mime_type));
-        }
 
         return $mime_type;
     }
@@ -844,6 +841,8 @@ class rcube_mime
             'image/jpg'      => array('jpg', 'jpeg', 'jpe'),
             'image/pjpeg'    => array('jpg', 'jpeg', 'jpe'),
             'image/tiff'     => array('tif'),
+            'image/bmp'      => array('bmp'),
+            'image/x-ms-bmp' => array('bmp'),
             'message/rfc822' => array('eml'),
             'text/x-mail'    => array('eml'),
         );
