@@ -3,7 +3,8 @@
 /**
  +-----------------------------------------------------------------------+
  | This file is part of the Roundcube Webmail client                     |
- | Copyright (C) 2008-2012, The Roundcube Dev Team                       |
+ |                                                                       |
+ | Copyright (C) The Roundcube Dev Team                                  |
  |                                                                       |
  | Licensed under the GNU General Public License version 3 or            |
  | any later version with exceptions for skins & plugins.                |
@@ -16,68 +17,32 @@
  | Author: Aleksander Machniak <alec@alec.pl>                            |
  | Author: Frederic Motte <fmotte@ubixis.com>                            |
  +-----------------------------------------------------------------------+
- */
 
-/*
- *                Washtml, a HTML sanityzer.
- *
- * Copyright (c) 2007 Frederic Motte <fmotte@ubixis.com>
- * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
- * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
- * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
- * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
- * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * OVERVIEW:
- *
- * Wahstml take an untrusted HTML and return a safe html string.
- *
- * SYNOPSIS:
- *
- * $washer = new washtml($config);
- * $washer->wash($html);
- * It return a sanityzed string of the $html parameter without html and head tags.
- * $html is a string containing the html code to wash.
- * $config is an array containing options:
- *   $config['allow_remote'] is a boolean to allow link to remote images.
- *   $config['blocked_src'] string with image-src to be used for blocked remote images
- *   $config['show_washed'] is a boolean to include washed out attributes as x-washed
- *   $config['cid_map'] is an array where cid urls index urls to replace them.
- *   $config['charset'] is a string containing the charset of the HTML document if it is not defined in it.
- * $washer->extlinks is a reference to a boolean that is set to true if remote images were removed. (FE: show remote images link)
- *
- * INTERNALS:
- *
- * Only tags and attributes in the static lists $html_elements and $html_attributes
- * are kept, inline styles are also filtered: all style identifiers matching
- * /[a-z\-]/i are allowed. Values matching colors, sizes, /[a-z\-]/i and safe
- * urls if allowed and cid urls if mapped are kept.
- *
- * Roundcube Changes:
- * - added $block_elements
- * - changed $ignore_elements behaviour
- * - added RFC2397 support
- * - base URL support
- * - invalid HTML comments removal before parsing
- * - "fixing" unitless CSS values for XHTML output
- * - SVG and MathML support
- */
+                Washtml, a HTML sanityzer.
+
+ Copyright (c) 2007 Frederic Motte <fmotte@ubixis.com>
+ All rights reserved.
+
+ Redistribution and use in source and binary forms, with or without
+ modification, are permitted provided that the following conditions
+ are met:
+ 1. Redistributions of source code must retain the above copyright
+    notice, this list of conditions and the following disclaimer.
+ 2. Redistributions in binary form must reproduce the above copyright
+    notice, this list of conditions and the following disclaimer in the
+    documentation and/or other materials provided with the distribution.
+
+ THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
+ IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
+ INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
+ NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
+ THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+*/
 
 /**
  * Utility class providing HTML sanityzer
@@ -87,7 +52,9 @@
  */
 class rcube_washtml
 {
-    /* Allowed HTML elements (default) */
+    /**
+     * @var array Allowed HTML elements (default)
+     */
     static $html_elements = array('a', 'abbr', 'acronym', 'address', 'area', 'b',
         'basefont', 'bdo', 'big', 'blockquote', 'br', 'caption', 'center',
         'cite', 'code', 'col', 'colgroup', 'dd', 'del', 'dfn', 'dir', 'div', 'dl',
@@ -122,19 +89,23 @@ class rcube_washtml
         'bvar', 'lowlimit', 'uplimit',
     );
 
-    /* Ignore these HTML tags and their content */
+    /**
+     * @var array Ignore these HTML tags and their content
+     */
     static $ignore_elements = array('script', 'applet', 'embed', 'object', 'style');
 
-    /* Allowed HTML attributes */
+    /**
+     * @var array Allowed HTML attributes
+     */
     static $html_attribs = array('name', 'class', 'title', 'alt', 'width', 'height',
         'align', 'nowrap', 'col', 'row', 'id', 'rowspan', 'colspan', 'cellspacing',
         'cellpadding', 'valign', 'bgcolor', 'color', 'border', 'bordercolorlight',
         'bordercolordark', 'face', 'marginwidth', 'marginheight', 'axis', 'border',
         'abbr', 'char', 'charoff', 'clear', 'compact', 'coords', 'vspace', 'hspace',
         'cellborder', 'size', 'lang', 'dir', 'usemap', 'shape', 'media',
-        'background', 'src', 'poster', 'href',
+        'background', 'src', 'poster', 'href', 'headers',
         // attributes of form elements
-        'type', 'rows', 'cols', 'disabled', 'readonly', 'checked', 'multiple', 'value',
+        'type', 'rows', 'cols', 'disabled', 'readonly', 'checked', 'multiple', 'value', 'for',
         // SVG
         'accent-height', 'accumulate', 'additive', 'alignment-baseline', 'alphabetic',
         'ascent', 'attributename', 'attributetype', 'azimuth', 'basefrequency', 'baseprofile',
@@ -174,7 +145,9 @@ class rcube_washtml
         'fontsize', 'fontweight', 'fontstyle', 'fontfamily', 'groupalign', 'edge', 'side',
     );
 
-    /* Elements which could be empty and be returned in short form (<tag />) */
+    /**
+     * @var array Elements which could be empty and be returned in short form (<tag />)
+     */
     static $void_elements = array('area', 'base', 'br', 'col', 'command', 'embed', 'hr',
         'img', 'input', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr',
         // MathML
@@ -182,35 +155,52 @@ class rcube_washtml
         'maligngroup', 'none', 'mprescripts',
     );
 
-    /* State for linked objects in HTML */
+    /** @var bool State indicating existence of linked objects in HTML */
     public $extlinks = false;
 
-    /* Current settings */
+    /** @var array Current settings */
     private $config = array();
 
-    /* Registered callback functions for tags */
+    /** @var array Registered callback functions for tags */
     private $handlers = array();
 
-    /* Allowed HTML elements */
+    /** @var array Allowed HTML elements */
     private $_html_elements = array();
 
-    /* Ignore these HTML tags but process their content */
+    /** @var array Ignore these HTML tags but process their content */
     private $_ignore_elements = array();
 
-    /* Elements which could be empty and be returned in short form (<tag />) */
+    /** @var array Elements which could be empty and be returned in short form (<tag />) */
     private $_void_elements = array();
 
-    /* Allowed HTML attributes */
+    /** @var array Allowed HTML attributes */
     private $_html_attribs = array();
 
-    /* Max nesting level */
+    /** @var string A prefix to be added to id/class/for attribute values */
+    private $_css_prefix;
+
+    /** @var int Max nesting level */
     private $max_nesting_level;
 
+    /** @var bool True if current document is XML */
     private $is_xml = false;
 
 
     /**
      * Class constructor
+     *
+     * @param array $p Configuration options:
+     *         allow_remote: is a boolean to allow link to remote resources (images/css)
+     *         blocked_src: string with image-src to be used for blocked remote images
+     *         show_washed: is a boolean to include washed out attributes as x-washed
+     *         cid_map: is an array where cid urls index urls to replace them
+     *         charset: is a string containing the charset of the HTML document,
+     *                  to be used if the charset is not defined in the document
+     *         css_prefix: A prefix to be added to id/class/for attribute values
+     *         html_elements: Additional allowed HTML elements
+     *         ignore_elements: Additional HTML elements to ignore
+     *         html_attribs: Additional allowed HTML attributes
+     *         void_elements: Elements which could be empty and be returned in short form (<tag />)
      */
     public function __construct($p = array())
     {
@@ -218,22 +208,30 @@ class rcube_washtml
         $this->_html_attribs    = array_flip((array)$p['html_attribs']) + array_flip(self::$html_attribs);
         $this->_ignore_elements = array_flip((array)$p['ignore_elements']) + array_flip(self::$ignore_elements);
         $this->_void_elements   = array_flip((array)$p['void_elements']) + array_flip(self::$void_elements);
+        $this->_css_prefix      = is_string($p['css_prefix']) && strlen($p['css_prefix']) ? $p['css_prefix'] : null;
 
-        unset($p['html_elements'], $p['html_attribs'], $p['ignore_elements'], $p['void_elements']);
+        unset($p['html_elements'], $p['html_attribs'], $p['ignore_elements'], $p['void_elements'], $p['css_prefix']);
 
         $this->config = $p + array('show_washed' => true, 'allow_remote' => false, 'cid_map' => array());
     }
 
     /**
      * Register a callback function for a certain tag
+     *
+     * @param string   $tag      HTML tag name
+     * @param callback $callback Callback function
      */
-    public function add_callback($tagName, $callback)
+    public function add_callback($tag, $callback)
     {
-        $this->handlers[$tagName] = $callback;
+        $this->handlers[$tag] = $callback;
     }
 
     /**
      * Check CSS style
+     *
+     * @param string $style CSS style
+     *
+     * @return string Washed CSS style
      */
     private function wash_style($style)
     {
@@ -255,7 +253,7 @@ class rcube_washtml
                     if (preg_match('/^url\(/i', $val)) {
                         if (preg_match('/^url\(\s*[\'"]?([^\'"\)]*)[\'"]?\s*\)/iu', $val, $match)) {
                             if ($url = $this->wash_uri($match[1])) {
-                                $value .= ' url(' . htmlspecialchars($url, ENT_QUOTES) . ')';
+                                $value .= ' url(' . htmlspecialchars($url, ENT_QUOTES, $this->config['charset']) . ')';
                             }
                         }
                     }
@@ -288,6 +286,10 @@ class rcube_washtml
 
     /**
      * Take a node and return allowed attributes and check values
+     *
+     * @param DOMNode $node Document element
+     *
+     * @return string Washed element attributes
      */
     private function wash_attribs($node)
     {
@@ -328,8 +330,9 @@ class rcube_washtml
                     if (preg_match('/^[a-z:]*url\(/i', $val)) {
                         if (preg_match('/^([a-z:]*url)\(\s*[\'"]?([^\'"\)]*)[\'"]?\s*\)/iu', $value, $match)) {
                             if ($url = $this->wash_uri($match[2])) {
-                                $result .= ' ' . $attr->nodeName . '="' . $match[1] . '(' . htmlspecialchars($url, ENT_QUOTES) . ')'
-                                     . substr($val, strlen($match[0])) . '"';
+                                $result .= ' ' . $attr->nodeName . '="' . $match[1]
+                                    . '(' . htmlspecialchars($url, ENT_QUOTES, $this->config['charset']) . ')'
+                                    . substr($val, strlen($match[0])) . '"';
                                 continue;
                             }
                         }
@@ -341,19 +344,22 @@ class rcube_washtml
                         $out = $value;
                     }
                 }
+                else if ($this->_css_prefix !== null && in_array($key, array('id', 'class', 'for'))) {
+                    $out = preg_replace('/(\S+)/', $this->_css_prefix . '\1', $value);
+                }
                 else if ($key) {
                    $out = $value;
                 }
 
                 if ($out !== null && $out !== '') {
-                    $result .= ' ' . $attr->nodeName . '="' . htmlspecialchars($out, ENT_QUOTES) . '"';
+                    $result .= ' ' . $attr->nodeName . '="' . htmlspecialchars($out, ENT_QUOTES | ENT_SUBSTITUTE, $this->config['charset']) . '"';
                 }
                 else if ($value) {
-                    $washed[] = htmlspecialchars($attr->nodeName, ENT_QUOTES);
+                    $washed[] = htmlspecialchars($attr->nodeName, ENT_QUOTES, $this->config['charset']);
                 }
             }
             else {
-                $washed[] = htmlspecialchars($attr->nodeName, ENT_QUOTES);
+                $washed[] = htmlspecialchars($attr->nodeName, ENT_QUOTES, $this->config['charset']);
             }
         }
 
@@ -366,8 +372,14 @@ class rcube_washtml
 
     /**
      * Wash URI value
+     *
+     * @param string $uri            URI
+     * @param bool   $blocked_source Block remote source
+     * @param bool   $is_image       URI points to an image
+     *
+     * @return string Washed URI
      */
-    private function wash_uri($uri, $blocked_source = false)
+    private function wash_uri($uri, $blocked_source = false, $is_image = true)
     {
         if (($src = $this->config['cid_map'][$uri])
             || ($src = $this->config['cid_map'][$this->config['base_url'].$uri])
@@ -386,17 +398,22 @@ class rcube_washtml
             }
 
             $this->extlinks = true;
-            if ($blocked_source && $this->config['blocked_src']) {
+            if ($is_image && $blocked_source && $this->config['blocked_src']) {
                 return $this->config['blocked_src'];
             }
         }
-        else if (preg_match('/^data:image.+/i', $uri)) { // RFC2397
+        else if ($is_image && preg_match('/^data:image.+/i', $uri)) { // RFC2397
             return $uri;
         }
     }
 
     /**
      * Check it the tag/attribute may contain an URI
+     *
+     * @param string $tag  Element name
+     * @param string $attr Attribute name
+     *
+     * @return bool True if attribute may contain an URI, False otherwise
      */
     private function is_link_attribute($tag, $attr)
     {
@@ -405,6 +422,11 @@ class rcube_washtml
 
     /**
      * Check it the tag/attribute may contain an image URI
+     *
+     * @param string $tag  Element name
+     * @param string $attr Attribute name
+     *
+     * @return bool True if attribute may contain an image URI, False otherwise
      */
     private function is_image_attribute($tag, $attr)
     {
@@ -417,6 +439,11 @@ class rcube_washtml
 
     /**
      * Check it the tag/attribute may contain a FUNCIRI value
+     *
+     * @param string $tag  Element name
+     * @param string $attr Attribute name
+     *
+     * @return bool True if attribute may contain a FUNCIRI value, False otherwise
      */
     private function is_funciri_attribute($tag, $attr)
     {
@@ -430,6 +457,8 @@ class rcube_washtml
      *
      * @param DOMNode $node  HTML element
      * @param int     $level Recurrence level (safe initial value found empirically)
+     *
+     * @return string HTML content
      */
     private function dumpHtml($node, $level = 20)
     {
@@ -448,6 +477,7 @@ class rcube_washtml
                     'message' => "Maximum nesting level exceeded (xdebug.max_nesting_level={$this->max_nesting_level})"),
                     true, false);
             }
+
             return '<!-- ignored -->';
         }
 
@@ -458,6 +488,17 @@ class rcube_washtml
             switch ($node->nodeType) {
             case XML_ELEMENT_NODE: //Check element
                 $tagName = strtolower($node->nodeName);
+
+                if ($tagName == 'link') {
+                    $uri = $this->wash_uri($node->getAttribute('href'), false, false);
+                    if (!$uri) {
+                        $dump .= '<!-- link ignored -->';
+                        break;
+                    }
+
+                    $node->setAttribute('href', (string) $uri);
+                }
+
                 if ($callback = $this->handlers[$tagName]) {
                     $dump .= call_user_func($callback, $tagName,
                         $this->wash_attribs($node), $this->dumpHtml($node, $level), $this);
@@ -475,7 +516,7 @@ class rcube_washtml
                         }
                     }
                     else if ($tagName == 'textarea' && strpos($content, '<') !== false) {
-                        $content = htmlspecialchars($content, ENT_QUOTES);
+                        $content = htmlspecialchars($content, ENT_QUOTES | ENT_SUBSTITUTE, $this->config['charset']);
                     }
 
                     $dump .= $this->wash_attribs($node);
@@ -488,10 +529,10 @@ class rcube_washtml
                     }
                 }
                 else if (isset($this->_ignore_elements[$tagName])) {
-                    $dump .= '<!-- ' . htmlspecialchars($node->nodeName, ENT_QUOTES) . ' not allowed -->';
+                    $dump .= '<!-- ' . htmlspecialchars($node->nodeName, ENT_QUOTES, $this->config['charset']) . ' not allowed -->';
                 }
                 else {
-                    $dump .= '<!-- ' . htmlspecialchars($node->nodeName, ENT_QUOTES) . ' ignored -->';
+                    $dump .= '<!-- ' . htmlspecialchars($node->nodeName, ENT_QUOTES, $this->config['charset']) . ' ignored -->';
                     $dump .= $this->dumpHtml($node, $level); // ignore tags not its content
                 }
                 break;
@@ -501,7 +542,7 @@ class rcube_washtml
                 break;
 
             case XML_TEXT_NODE:
-                $dump .= htmlspecialchars($node->nodeValue);
+                $dump .= htmlspecialchars($node->nodeValue, ENT_COMPAT | ENT_HTML401 | ENT_SUBSTITUTE, $this->config['charset']);
                 break;
 
             case XML_HTML_DOCUMENT_NODE:
@@ -517,11 +558,13 @@ class rcube_washtml
     /**
      * Main function, give it untrusted HTML, tell it if you allow loading
      * remote images and give it a map to convert "cid:" urls.
+     *
+     * @param string $html HTML content
+     *
+     * @return string Washed HTML content
      */
     public function wash($html)
     {
-        // Charset seems to be ignored (probably if defined in the HTML document)
-        $node = new DOMDocument('1.0', $this->config['charset']);
         $this->extlinks = false;
 
         $html = $this->cleanup($html);
@@ -540,15 +583,22 @@ class rcube_washtml
         // SVG need to be parsed as XML
         $this->is_xml = stripos($html, '<html') === false && stripos($html, '<svg') !== false;
         $method       = $this->is_xml ? 'loadXML' : 'loadHTML';
-        $options      = 0;
 
-        // Use optimizations if supported
-        if (PHP_VERSION_ID >= 50400) {
-            $options = LIBXML_PARSEHUGE | LIBXML_COMPACT | LIBXML_NONET;
-            @$node->{$method}($html, $options);
+        // DOMDocument does not support HTML5, try Masterminds parser if available
+        if (!$this->is_xml && class_exists('Masterminds\HTML5')) {
+            try {
+                $html5 = new Masterminds\HTML5();
+                $node  = $html5->loadHTML($this->fix_html5($html));
+            }
+            catch (Exception $e) {
+                // ignore, fallback to DOMDocument
+            }
         }
-        else {
-            @$node->{$method}($html);
+
+        if (empty($node)) {
+            // Charset seems to be ignored (probably if defined in the HTML document)
+            $node = new DOMDocument('1.0', $this->config['charset']);
+            @$node->{$method}($html, LIBXML_PARSEHUGE | LIBXML_COMPACT | LIBXML_NONET);
         }
 
         return $this->dumpHtml($node);
@@ -556,6 +606,10 @@ class rcube_washtml
 
     /**
      * Getter for config parameters
+     *
+     * @param string $prop Configuration parameter name
+     *
+     * @return mixed Configuration parameter value
      */
     public function get_config($prop)
     {
@@ -564,6 +618,10 @@ class rcube_washtml
 
     /**
      * Clean HTML input
+     *
+     * @param string $html HTML content
+     *
+     * @return string Clean HTML content
      */
     private function cleanup($html)
     {
@@ -581,6 +639,8 @@ class rcube_washtml
             '/^(\0\0\xFE\xFF|\xFF\xFE\0\0|\xFE\xFF|\xFF\xFE|\xEF\xBB\xBF)/',
             // washtml/DOMDocument cannot handle xml namespaces
             '/<html\s[^>]+>/i',
+            // washtml/DOMDocument cannot handle xml namespaces
+            '/<\?xml:namespace\s[^>]+>/i',
         );
 
         $html_replace = array(
@@ -589,9 +649,15 @@ class rcube_washtml
             '',
             '',
             '<html>',
+            '',
         );
 
-        $html = preg_replace($html_search, $html_replace, trim($html));
+        $html = preg_replace($html_search, $html_replace, $html);
+
+        $err = array('line' => __LINE__, 'file' => __FILE__, 'message' => "Could not clean up HTML!");
+        if ($html === null && rcube_utils::preg_error($err)) {
+            return '';
+        }
 
         // Replace all of those weird MS Word quotes and other high characters
         $badwordchars = array(
@@ -614,24 +680,6 @@ class rcube_washtml
 
         $html = str_replace($badwordchars, $fixedwordchars, $html);
 
-        // PCRE errors handling (#1486856), should we use something like for every preg_* use?
-        if ($html === null && ($preg_error = preg_last_error()) != PREG_NO_ERROR) {
-            $errstr = "Could not clean up HTML message! PCRE Error: $preg_error.";
-
-            if ($preg_error == PREG_BACKTRACK_LIMIT_ERROR) {
-                $errstr .= " Consider raising pcre.backtrack_limit!";
-            }
-            if ($preg_error == PREG_RECURSION_LIMIT_ERROR) {
-                $errstr .= " Consider raising pcre.recursion_limit!";
-            }
-
-            rcube::raise_error(array('code' => 620, 'type' => 'php',
-                'line' => __LINE__, 'file' => __FILE__,
-                'message' => $errstr), true, false);
-
-            return '';
-        }
-
         // FIXME: HTML comments handling could be better. The code below can break comments (#6464),
         //        we should probably do not modify content inside comments at all.
 
@@ -654,6 +702,10 @@ class rcube_washtml
 
     /**
      * Callback function for HTML tags fixing
+     *
+     * @param array $matches Matched elements (from preg_replace_callback())
+     *
+     * @return string Replacement string
      */
     public static function html_tag_callback($matches)
     {
@@ -679,6 +731,10 @@ class rcube_washtml
 
     /**
      * Convert all relative URLs according to a <base> in HTML
+     *
+     * @param string $body HTML body
+     *
+     * @return string HTML body
      */
     public static function resolve_base($body)
     {
@@ -693,6 +749,8 @@ class rcube_washtml
 
     /**
      * Fix broken nested lists, they are not handled properly by DOMDocument (#1488768)
+     *
+     * @param string &$html HTML content
      */
     public static function fix_broken_lists(&$html)
     {
@@ -722,8 +780,8 @@ class rcube_washtml
                     // li close tag
                     else if ($tt == '</li' && in_array($html[$p+4], array(' ', '>'))) {
                         $li_pos = $p;
+                        $in_li  = false;
                         $p += 4;
-                        $in_li = false;
                     }
                     // ul/ol closing tag
                     else if ($tt == '</' . $tag && in_array($html[$p+4], array(' ', '>'))) {
@@ -751,8 +809,8 @@ class rcube_washtml
                         $element = substr($html, $p, $len);
 
                         // move element to the end of the last li
-                        $html    = substr_replace($html, '', $p, $len);
-                        $html    = substr_replace($html, $element, $li_pos, 0);
+                        $html = substr_replace($html, '', $p, $len);
+                        $html = substr_replace($html, $element, $li_pos, 0);
 
                         $p = $end;
                     }
@@ -764,8 +822,49 @@ class rcube_washtml
         }
     }
 
+    /**
+     * Cleanup and workarounds on input to Masterminds/HTML5
+     *
+     * @param string $html HTML content
+     *
+     * @return string HTML content
+     */
+    protected function fix_html5($html)
+    {
+        // There might be content before html/body tag, we'll move it to the body
+        // We'll wrap it by a div container, it's an invalid HTML anyway
+        if (strpos($html, '<')) {
+            $pos     = stripos($html, '<!DOCTYPE') ?: stripos($html, '<html') ?: stripos($html, '<body');
+            $prefix  = '<div>' . substr($html, 0, $pos) . '</div>';
+            $html    = substr($html, $pos);
+        }
+
+        // HTML5 requires <head> or <body> (#6713)
+        // https://github.com/Masterminds/html5-php/issues/166
+        if (isset($prefix) || !preg_match('/<(head|body)/i', $html)) {
+            $body_pos = stripos($html, '<body');
+            $pos      = $body_pos !== false ? $body_pos : stripos($html, '<html');
+
+            // No HTML and no BODY tag
+            if ($pos === false) {
+                $html = '<html><body>' . $prefix . $html;
+            }
+            // Either HTML or BODY tag found
+            else {
+                $pos  = strpos($html, '>', $pos);
+                $html = substr_replace($html, ($body_pos === false ? '<body>' : '') . $prefix, $pos + 1, 0);
+            }
+        }
+
+        return $html;
+    }
+
     /**
      * Explode css style value
+     *
+     * @param string $style CSS style
+     *
+     * @return array List of CSS rules
      */
     protected function explode_style($style)
     {
